/**
  (c) Copyright CERN 2005. All rights not expressly granted are reserved.
  icecontrols.support@cern.ch

  SPDX-License-Identifier: LGPL-3.0-only
**/

#uses "classes/unTree/delegates/unTreeDelegateFactory.ctl"
#uses "classes/fwTree/FwTree_SimpleImpl.ctl"
#uses "classes/unTree/unTreeRepositoryImpl.ctl"
#uses "classes/unTree/unTreeProcessor.ctl"

class UnTreeWidgetData{
    protected FwTreeNodePtr _selectedNode = nullptr;
    protected shared_ptr<FwTreeNodeDelegate> _delegate = nullptr;
    protected FwTreeNodePtr _root = nullptr;
    protected shared_ptr<UnTree_RepositorySimple> _repo = nullptr;
    protected shape _widget;

    public UnTreeWidgetData(const string treeType, const string widgetType, const shape widget){
        _widget = widget;
        this.regenerate(treeType,widgetType);
        uiConnect(this, "selectionChangedCB", widget, "clicked");
    }

    public selectionChangedCB(FwTreeNodePtr node){
        _selectedNode = node;
    }

    public void regenerate(const string treeType, const string widgetType){
        assignPtr(_delegate, UnTreeDelegateFactory::getDelegate(treeType,widgetType));
        if(widgetType == TREE_WIDGET_PREVIEW || widgetType == TREE_WIDGET_FILE){
            assignPtr(
                _repo,
                widgetType == TREE_WIDGET_PREVIEW ?
                UnTree_RepositorySimple::getInstance() :
                UnTree_RepositoryReadImpl::getInstance()
            );
            assignPtr(_root,_repo.createRootNode(treeType+"Tree"+widgetType));
            // _root.addChild("---Clipboard");
            invokeMethod(_widget,"initializeTree",_delegate,"",_repo);
            // EditingTree.initializeTree(myTNDelegate,"",myTreeRepo);
        } else {
            invokeMethod(_widget,"initializeTree",_delegate,treeType+"Tree");
            // invokeMethod(_widget,"setTreeType",UnTreeDelegateFactory::getTreeTypeDP(treeType,widgetType));
            // invokeMethod(_widget,"setTreeItemDelegate",_delegate);
            const string treeName = "fwTN_" + treeType + "Tree";
            _root = FwTree_Repository::getDefault().get(treeName);//.getRootNode();
        }
        assignPtr(_selectedNode,nullptr);
    }

    public void removeSelectedNode(){
        if(equalPtr(_selectedNode,nullptr)){
            return;
        }
        _selectedNode.removeMe(true);
        _selectedNode = nullptr;
    }

    public shared_ptr<FwTreeNodeDelegate> getDelegate() { return _delegate; }
    public FwTreeNodePtr getRootNode() { return _root; }
    public FwTreeNodePtr getSelectedNode() { return _selectedNode; }
    public shape getWidget() { return _widget; }

};

class UnTreeMultitreeHandler{
    #event logMessage(string msg, string type = "INFO")
    protected mapping _data; // mapping - keys are widget types
    protected string _treeType; //window or trend

    public UnTreeMultitreeHandler(
        const string treeType,
        mapping widgets // tree ids and shapes
    ){
        _treeType = treeType;
        this._initializeTrees(mappingKeys(widgets));
        for(int i=0; i<widgets.count(); ++i){
            _data.insert( widgets.keyAt(i), new UnTreeWidgetData(_treeType, widgets.keyAt(i), widgets.valueAt(i)) );
        }
        _updateRelations();
        if(widgets.contains(TREE_WIDGET_PREVIEW)){
            invokeMethod(widgets[TREE_WIDGET_PREVIEW],"setDragAndDropEnabled",false,true);
            // invokeMethod(widgets[TREE_WIDGET_PREVIEW],"setEditable",true);
            // this.clearTree(TREE_WIDGET_PREVIEW);
        }
        if(widgets.contains(TREE_WIDGET_CURRENT)){
            invokeMethod(widgets[TREE_WIDGET_CURRENT],"setDragAndDropEnabled",true,false);
            // invokeMethod(widgets[TREE_WIDGET_CURRENT],"setEditable",true);
        }

        if(widgets.contains(TREE_WIDGET_FILE)){
            invokeMethod(widgets[TREE_WIDGET_FILE],"setDragAndDropEnabled",true,false);
            // invokeMethod(widgets[TREE_WIDGET_FILE],"setEditable",true);
            // this.clearTree(TREE_WIDGET_FILE);
        }
    }

    public string getTreeType() { return this._treeType; }

    public bool areNodesSelected(dyn_string names = makeDynString()){
        if(dynlen(names) == 0){
            names = mappingKeys(_data);
        }
        for(int i=0; i<dynlen(names); ++i){
            if(!_data.contains(names.at(i))){
                return false;
            }
            if(equalPtr(_data[names.at(i)].getSelectedNode(), nullptr)){
                return false;
            }
        }
        return true;
    }

    public bool areTreesDefined(dyn_string names = makeDynString()){
        if(dynlen(names) == 0){
            names = mappingKeys(_data);
        }
        for(int i=0; i<dynlen(names); ++i){
            if(!_data.contains(names.at(i))){
                return false;
            }
            if(equalPtr(_data[names.at(i)].getRootNode(), nullptr)){
                return false;
            }
        }
        return true;
    }

    protected void regenerate(){
        this._initializeTrees(mappingKeys(_data));
        for(int i=0; i<_data.count(); ++i){
            _data.valueAt(i).regenerate(_treeType,_data.keyAt(i));
        }
        _updateRelations();
    }

    protected void _updateRelations(){
        if(_data.contains(TREE_WIDGET_PREVIEW)){
            if(_data.contains(TREE_WIDGET_FILE)){
                _data[TREE_WIDGET_PREVIEW].getDelegate().addSourceRepo(
                    "file",
                    _data[TREE_WIDGET_FILE].getRootNode().getRepo()
                );
            }
        }
    }

    // public setColorMapping(const string widgetType, const mapping map){
    //     FwException::assert(_data.contains(widgetType)," Nonexising widget type: " + widgetType);
    //     _data[widgetType].getDelegate().setColorMapping(map);
    //     invokeMethod(_data[widgetType].getWidget(),"refresh");
    //     invokeMethod(_data[widgetType].getWidget(),"expandAll");
    // }

    public FwTreeNodePtr getRootNode(const string source){
        if(source == TREE_WIDGET_CURRENT){
            const string treeName = "fwTN_" + _treeType + "Tree";
            return FwTree_Repository::getDefault().get(treeName);
        }
        if(!_data.contains(source)) return nullptr;
        return _data[source].getRootNode();
    }

    public setDensity(unsigned newDensity){
        for(int i=0; i<_data.count(); ++i){
            _data.valueAt(i).getDelegate().setUiDensity(newDensity);
            invokeMethod(_data.valueAt(i).getWidget(),"setUiDensity",newDensity);
            invokeMethod(_data.valueAt(i).getWidget(),"refresh");
        }
    }

    public void changeTreeType(const string newType){
        if(newType == _treeType){
            return;
        }
        _treeType = newType;
        this.regenerate();
        return;
    }

    public void clearTree(const string widgetType, const bool withClipboard = true){
        if(!_data.contains(widgetType)){ return; }
        if( equalPtr(_data[widgetType].getRootNode(), nullptr) ){ return; }
        if(withClipboard){
            FwTreeNodePtr clipboard = _data[widgetType].getRootNode().getClipboard();
            if(!equalPtr(clipboard,nullptr)){
                this._clearNodes(clipboard.getChildren());
            }
        }
        this._clearNodes(_data[widgetType].getRootNode().getChildren());
    }

    public void mergeTrees(string source1, string source2, string target){
        FwTreeNodePtr s2 = getRootNode(source2);
        FwTreeNodePtr tnTarget = getRootNode(target);
        this._clearNodes(tnTarget.getChildren());
        copyBetweenTrees(source1,target,"all");
        UnTreeProcessor::mergeNodes(s2,tnTarget);
    }

    public void copyBetweenTrees(const string source, const string target, const string location){
        if(source != TREE_WIDGET_CURRENT || location!="all"){
            if(!this.areTreesDefined(source,target)) { return; }
        }
        if(location=="all"){
            // this.clearTree(target);
            FwTreeNodePVec kids;
            if(source == TREE_WIDGET_CURRENT && location == "all"){
                const string treeName = "fwTN_" + _treeType + "Tree";
                FwTreeNodePtr topNode = FwTree_Repository::getDefault().get(treeName);//.getRootNode();
                kids =  topNode.getChildren();
            } else {
                kids = _data[source].getRootNode().getChildren();
            }
            for(int i=0; i<kids.count(); ++i){
                logMessage("COPYING NODE: " + kids.at(i).getName() + "[" + kids.at(i).getDP() + "]");
                FwTreeNodeDelegate::copyNodeStructure(_data[target].getRootNode(), kids.at(i));
            }
            invokeMethod(_data[target].getWidget(),"refresh");
            //UNCORE-433 for performance reason, expand only the first node
            invokeMethod(_data[target].getWidget(),"expandNode", _data[target].getRootNode());
            return;

        }
        if(!this.areNodesSelected(source,target)) { return; }
        FwTreeNodePtr srcNode = _data[source].getSelectedNode();
        FwTreeNodePtr dstNode = _data[target].getSelectedNode();

        logMessage("COPYING NODE: " + srcNode.getName() + "[" + srcNode.getDP() + "]");
        if(location == "below") FwTreeNodeDelegate::copyNodeStructure(dstNode, srcNode);
        else if(location == "before") _data[target].getDelegate().handleExternalNodeDroppedBetween(dstNode, srcNode, true);
        else if(location == "after") _data[target].getDelegate().handleExternalNodeDroppedBetween(dstNode, srcNode, false);
    }

    public void removeSelectedNode( const string widgetType){
        if(_data.contains(widgetType)){

        }
        _data[widgetType].removeSelectedNode();
    }
    protected void _clearNodes(FwTreeNodePVec nodes){
        for(int i=0; i<nodes.count(); ++i){
            logMessage("REMOVING NODE: " + nodes.at(i).getName() + "<" + nodes.at(i).getDP() + ">");
            nodes[i].removeMe(true);
        }
    }

    protected void _initializeTrees(dyn_string types){
        dyn_string exceptionInfo;
        dyn_string allRoots;
        fwTree_getRootNodes(allRoots,exceptionInfo);
        FwException::checkRaise(exceptionInfo);

        for(int i=0; i<dynlen(types); ++i){
            const string treeName = UnTreeDelegateFactory::getTreeTypeDP(_treeType,types.at(i));
            if(!allRoots.contains(treeName)){
                fwTreeDisplay_createTree(treeName, 2, treeName, makeDynString(), makeDynString());
            }
        }
    }

};
