/**
  (c) Copyright CERN 2005. All rights not expressly granted are reserved.
  icecontrols.support@cern.ch

  SPDX-License-Identifier: LGPL-3.0-only
**/

#uses "classes/fwTree/FwTree.ctl"
#uses "classes/unTree/unTreeProcessContext.ctl"
#uses "classes/unTree/fileParsers/unTreeFileParser.ctl"
#uses "classes/fwTree/FwTree_RepositoryImpl.ctl"
// #uses "classes/unTree/deviceParsers/unTreeDeviceParser.ctl"
#uses "classes/unTree/deviceParsers/unTreeDeviceParserFactory.ctl"

//assuming device name -> actual value
//node name -> just name displayed in tree

class UnTreeNodeParser {
    protected shared_ptr<UnTreeProcessContext> _context = nullptr;
    protected FwTreeNodePtr _node = nullptr;
    protected shared_ptr<mapping> _data = nullptr;
    protected shared_ptr<UnTreeDeviceParser> _device = nullptr;

    public shared_ptr<UnTreeDeviceParser> getDeviceParser(){ return _device; }

    protected void setStatus(const int status){
        _context.setStatus(this.getParamByName(NODE_PARAMNAME_NAME),status);
    }

    protected void assert(bool cond, const string msg){
        if(cond){ return; }
        this.setStatus(NODE_STATUS_ERR);
        FwException::raise(msg);
    }

    public FwTreeNodePtr getNode() { return _node; }

    public UnTreeNodeParser(FwTreeNodePtr node = nullptr, shared_ptr<UnTreeProcessContext> context = nullptr){
        _context = context;
        if(_context == nullptr){
            _context = new UnTreeProcessContext();
        }
        _node = node;
    }

    // public bool hasHierarhyDefined(){
    //     if(!equalPtr(_node,nullptr)){ return true; }
    //     if(!equalPtr(_data,nullptr)){ return false; }
    //     if(!_data.contains(NODE_PARAMNAME_PARENT)){ return false; }
    //     if(!_data.contains(NODE_PARAMNAME_CHILDREN)){ return false; }
    // }

    public bool hasDevice(){
        if(!equalPtr(_node,nullptr)){ return _node.getDevice() != ""; }
        return !equalPtr(_device,nullptr);
    }

    public void createFromMap(FwTreeNodePtr parent, shared_ptr<mapping> data){
        assignPtr(_data,data);
        assignPtr(_node,nullptr);
        assignPtr(_device,nullptr);
        this.assert(!equalPtr(parent,nullptr),"Parent cannot be null");
        this.assert(data.value(NODE_PARAMNAME_NAME,"") != "", "Could not import node without a name");
        string name=data[NODE_PARAMNAME_NAME];
        if(name.startsWith("fwTN_") || name.startsWith("unTN_")) name=substr(name, 5); // cut the leading "fwTN_"
        // if(name.startsWith("&")) name=substr(name, 5); // cut the &0001
        _node = parent.getRepo().create(name,parent);
        if(data.contains(NODE_PARAMNAME_DEVICEDATA)){
            const dyn_string deviceData = UnTreeFileParser::decodeList(data[NODE_PARAMNAME_DEVICEDATA],UN_PARAMETER_DELIMITER,3);
            const string linkedDevice = _context.decodeDeviceName( deviceData.at(0) );
            const string type = deviceData.at(1);
            const bool isMaster = deviceData.at(2)=="1";
            if(type != ""){
                _device = UnTreeDeviceParserFactory::createDeviceParser(linkedDevice,_context,type);
                ;
                shared_ptr<mapping> devData = data[NODE_PARAMNAME_DEVICECONFIGURATION][type];
                _device.parseFromMap(devData);
//                 if(getType(data[NODE_PARAMNAME_USERDATA]) == DYN_STRING_VAR)
                dyn_string userData = UnTreeFileParser::decodeList(data[NODE_PARAMNAME_USERDATA],c_panel_navigation_delimiter);
                _node.setProps(type, linkedDevice, userData, isMaster);
                // if(_device.hasExistingDp()){
                //     this.setStatus(_device.matchesWithDP() ? NODE_STATUS_OK : NODE_STATUS_DIFF);
                // } else {
                //     this.setStatus(NODE_STATUS_NEW);
                // }
            }
        } else {
            this.setStatus(NODE_STATUS_N_A);
        }
    }

    public void fixChildrenOrder(){
        FwException::assertNotNull(_node);
        FwException::assertNotNull(_data);
        FwTreeNodePVec children;

        const dyn_string childrenNames = UnTreeFileParser::decodeList(_data.value(NODE_PARAMNAME_CHILDREN,""),UN_PARAMETER_DELIMITER);
        for(int i=0; i<dynlen(childrenNames); ++i){
            string name = childrenNames.at(i);
            if(name.startsWith("fwTN_") || name.startsWith("unTN_")) name=substr(name, 5); // cut the leading "fwTN_"
            FwTreeNodePtr child = _node.getRepo().get(name);
            this.assert(!equalPtr(child,nullptr),"Could not find child:" + childrenNames.at(i));
            children.append(child);
        }
        _node.reorderChildren(children);
    }

    public void applyHierarhy(){
        FwException::assertNotNull(_node);
        FwException::assertNotNull(_data);
        string name = _data.value(NODE_PARAMNAME_PARENT,"");
        if(name.startsWith("fwTN_") || name.startsWith("unTN_")) name=substr(name, 5); // cut the leading "fwTN_"
        if(name.startsWith("WindowTree") || name == "WindowTree" || name == ""){
            name = _node.getRootNode().getDP();
        }
        else if(name.startsWith("TrendTree") || name == "TrendTree" || name == ""){
            name = _node.getRootNode().getDP();
        }
        // this.assert(name!="","Node without parent");
        FwTreeNodePtr parent = _node.getRepo().get(name);
        this.assert(!equalPtr(parent,nullptr),"Could not find parent: " + name);
        _node.move(parent);
    }

    public mixed getParamByName(const string key, const anytype defVal = nullptr){
        if(equalPtr(_node,nullptr)){
            if(equalPtr(_data,nullptr)){ return defVal; }
            return _data.value(key,defVal);
        }
        switch(key){
            case NODE_PARAMNAME_NAME:
                return _context.encodeNodeName(_node.getDP());
            case NODE_PARAMNAME_PARENT:{
                FwTreeNodePtr parent = _node.getParent();
                return equalPtr(parent,nullptr) ? "" : _context.encodeNodeName(parent.getDP());
            }
            case NODE_PARAMNAME_CHILDREN:
                return getChildrenNamesValue();
            case NODE_PARAMNAME_USERDATA:
                return _node.getUserData();
            case NODE_PARAMNAME_DEVICEDATA:
                return this.getDeviceDataValue();
            default:
                return defVal;
        }
    }

    public void exportParam(shared_ptr<UnTreeFileParser> fileParser, const string key){
        fileParser.addValue(key,this.getParamByName(key));
    }

    public dyn_string getChildrenNamesValue(){
        dyn_string result;
        FwTreeNodePVec children = _node.getChildren();
        for ( int i=0 ; i<children.count() ; ++i ) {
            result.append(
                _context.encodeNodeName( children.at(i).getDP() )
            );
        }
        return result;
    }

    public dyn_string getDeviceDataValue(){
        string type = "";
        const string device = _node.getLinkedObj();
        if(device != "") {
            if(!dpExists(device)){
                _context.throwIfNeeded( "Linked device does not exist: " + device );
            } else {
                type = dpTypeName(device);
            }
        }
        return makeDynString(
            _context.encodeDeviceName(device),
            _context.encodeDeviceName(type),
            _node.getIsCU() ? "1" : "0"
        );
    }

    public void exportToFile(shared_ptr<UnTreeFileParser> fileParser, const string externalParent = ""){
        // _context.addDevice(UnTreeProcessorParam::EXPORTED_DEVICES,this.getParamByName(NODE_PARAMNAME_NAME));
        anytype mainNode = fileParser.addNode(unTreeWidget_treeNodeDPT, fileParser.getTopNode());
        this.exportParam(fileParser, NODE_PARAMNAME_NAME);
        if(!_context.get(UnTreeProcessorParam::SKIP_TREE_DEF,false)){
            if(externalParent == "")
                this.exportParam(fileParser, NODE_PARAMNAME_PARENT);
            else
                fileParser.addValue(NODE_PARAMNAME_PARENT,externalParent);
        }
        this.exportParam(fileParser, NODE_PARAMNAME_CHILDREN);
        this.exportParam(fileParser, NODE_PARAMNAME_DEVICEDATA);
        const string linkedDevice = _node.getLinkedObj();
        if(linkedDevice == ""){
            fileParser.addValue(NODE_PARAMNAME_USERDATA, _node.getUserData(),-1,"~");
            return;
        }

        shared_ptr<UnTreeDeviceParser> deviceParser = UnTreeDeviceParserFactory::createDeviceParser(linkedDevice,_context);
        fileParser.addValue(
            NODE_PARAMNAME_USERDATA,
            deviceParser.encodeUserData(_node.getUserData()),
            -1,
            "~");
        const anytype deviceConfigurationNode = fileParser.addNode(NODE_PARAMNAME_DEVICECONFIGURATION,mainNode);
        try{
            deviceParser.exportToFile(fileParser);
        } catch {
            FwException::rethrow();
        } finally {
            fileParser.removeIfEmpty(deviceConfigurationNode);
        }
    }

};
