/**
  (c) Copyright CERN 2005. All rights not expressly granted are reserved.
  icecontrols.support@cern.ch

  SPDX-License-Identifier: LGPL-3.0-only
**/

#uses "classes/unTree/fileParsers/unTreeFileParser.ctl"

class UnTreeFileParserXml:UnTreeFileParser{
    protected int _xmlDoc=-1;
    protected shared_ptr<mapping> _parsed;

    public UnTreeFileParserXml(shared_ptr<UnTreeProcessContext> context = nullptr):UnTreeFileParser(context){
        _parsed = new mapping();
    }

    public shared_ptr<mapping> asMap() { return _parsed; }


    public initializeExportFile(const string filePath, const FwTreeNodePtr node){
        _path = filePath;
        string details;
        if(!this.checkFileAccessExport(filePath,details)){
            FwException::raise(details);
        }
        FwTreeNodePtr rootNode = node.isRoot() ? node : node.getRootNode();
        _xmlDoc = xmlNewDocument();
        _rootNode = xmlAppendChild(_xmlDoc,-1,XML_PROCESSING_INSTRUCTION_NODE,"xml version=\"1.0\" encoding=\"UTF-8\"");
        _rootNode = xmlAppendChild(_xmlDoc,-1,XML_ELEMENT_NODE,"tree");
        _currentTop = _rootNode;
        xmlSetElementAttribute(_xmlDoc,_rootNode,"version",_context.get(UnTreeProcessorParam::FORMAT_VERSION));

        const string xmlName = _context.checkValue(UnTreeProcessorParam::TREE_NAME,"",rootNode.getName());
        // unExportTree_uniqueTreeTagNames();
        this.addValue("name",xmlName);
    }

    public int getFirstSiblingElement(int node){
        while ( xmlNodeType(_xmlDoc,node) != XML_ELEMENT_NODE ) {
            node = xmlNextSibling(_xmlDoc,node);
            if(node == -1){
                break;
            }
        }
        return node;
    }

    void assertCond(const bool condition, const string msg){
        if(condition){
            return;
        }
        if(_xmlDoc > 0){
            xmlCloseDocument(_xmlDoc);
        }
        FwException::raise(msg);
    }

    public void initializeImportFile(const string filePath){
        _path = filePath;
        string details;
        if(!this.checkFileAccessImport(filePath,details)){
            FwException::raise(details);
        }
        int errLine, errColumn;
        string errMsg;
        _xmlDoc = xmlDocumentFromFile(filePath,errMsg,errLine,errColumn);
        this.assertCond(_xmlDoc >=0 ,"XML error: '" + errMsg + "' at line " + errLine + " column " + errColumn );
        _rootNode = xmlFirstChild(_xmlDoc);
        this.assertCond(_xmlDoc >=0 , "XML content wrong - missing top node");
        _rootNode = this.getFirstSiblingElement(_rootNode);
        this.assertCond(_xmlDoc >=0 , "XML content wrong - missing top element-type node");
        const string topNodeName = xmlNodeName(_xmlDoc,_rootNode);
        this.assertCond(topNodeName == "tree", "XML file must start with element-node 'tree'." );
        string versionNumber;
        int res = xmlGetElementAttribute(_xmlDoc,_rootNode,"version",versionNumber);
        this.assertCond(res==0, "XML old format (no version tag found)");
        this.assertCond(versionNumber=="2","XML version NOT supported: " + versionNumber);
        dyn_int xmlEls;
        res = xmlChildNodes(_xmlDoc,_rootNode,xmlEls);
        this.assertCond(res==0 && dynlen(xmlEls) > 0, "XML file does not contain 'tree' children.");
    }

    protected void _insert(mixed parent, const string key, mixed element){
        if(getType(parent) == MAPPING_VAR){
            if(mappingHasKey(parent, key)){
                if(getType(parent[key]) != DYN_MIXED_VAR){
                    mixed temp = parent[key];
                    parent.remove(key);
                    parent.insert(key, new dyn_mixed());
                    parent[key].append(temp);
                } 
                parent[key].append(element);
            } else {
                parent[key] = element;
            }
        } else if ( getTypeName(parent).startsWith("dyn_") ){
            parent.append(element);
        } else {
            FwException::raise("unsupported parent type: " + getTypeName(parent) + ", key:", + key);
        }
    }

    public void parseFromFile(const string filePath){
        this.initializeImportFile(filePath);
        this.parseNodeFromFile(_rootNode,_parsed);
        xmlCloseDocument(_xmlDoc);
    }

    protected void parseNodeFromFile(const int nodeId, mixed parentElement){
        // _parsed.clear();
        const string nodeName = xmlNodeName(_xmlDoc,nodeId);
        dyn_int childrenIds, grandChildren;
        int res = xmlChildNodes(_xmlDoc,nodeId,childrenIds);
        for ( int i=1 ; i<=dynlen(childrenIds) ;++i ){
            const string name = xmlNodeName(_xmlDoc,childrenIds[i]);
            const int actualNode = xmlFirstChild(_xmlDoc,childrenIds[i]);
            if(actualNode < 0){
                this._insert(parentElement,name,"");
                continue;
            }
            const int nodeType = xmlNodeType(_xmlDoc,actualNode);
            if ( nodeType == XML_TEXT_NODE ) {
                this._insert(parentElement,name,(string)xmlNodeValue(_xmlDoc,actualNode));

            } else if ( nodeType == XML_ELEMENT_NODE ) {
            // } else if ( xmlChildNodes(_xmlDoc,childrenIds[i],grandChildren) > 0 && dynlen(grandChildren) > 0 ) {
                shared_ptr<mapping> newMap = new mapping();
                this._insert(parentElement,name,newMap);
                this.parseNodeFromFile(childrenIds[i],newMap);
            } else {
                this._insert(parentElement,name,(string)xmlNodeValue(_xmlDoc,actualNode));
                // FwException::raise("Unsupported node type encounterd: " + nodeType);
            }
		}
    }

    public mixed addNode(const string key, mixed parent=-1, const bool setAsTop = true){
        if(getType(parent)==INT_VAR && parent == -1){
            parent = _currentTop;
        }
        const int child = xmlAppendChild(_xmlDoc,parent,XML_ELEMENT_NODE,key);
        if(setAsTop){
            _currentTop = child;
        }
        return child;
    }

    public void resetTop(anytype topNode = -1) {
        if(getType(topNode)==INT_VAR && topNode == -1){
            _currentTop = _rootNode;
        } else {
            _currentTop = topNode;
        }
    }

    public void addValue(const string key, const anytype value, mixed parent = -1, const string sep = UN_PARAMETER_DELIMITER){
        if(value == "" && _context.get(UnTreeProcessorParam::SPARSE)){
            return;
        }
        if(getType(parent)==INT_VAR && parent == -1){
            parent = _currentTop;
        }
        const int child = this.addNode(key,parent,false);
        if(getTypeName(value).startsWith("dyn_")){
            if(sep == ""){
                for(int i=0; i<dynlen(value); ++i){
                    this.addValue(key,value.at(i),child);
                }
            } else {
                xmlAppendChild(_xmlDoc,child,XML_TEXT_NODE, strjoin(value, sep) );
            }
        } else if(getType(value) == MAPPING_VAR){
            for(int i=0; i<value.count(); ++i){
                this.addValue(value.keyAt(i),value.valueAt(i),child,sep);
            }
        } else {
            xmlAppendChild(_xmlDoc,child,XML_TEXT_NODE, value );
        }
    }

    public void writeToFile(){
        bool result = false;
        string tempPath = _path;
        if(_context.get(UnTreeProcessorParam::BEAUTIFY)){
            if(_path.endsWith(".txt")){
                strreplace(tempPath,".txt","_temp.txt");
            } else if(_path.endsWith(".xml")){
                strreplace(tempPath,".xml","_temp.xml");
            } else {
                tempPath = tempPath + "_temp";
            }
            result = xmlDocumentToFile(_xmlDoc,tempPath) != -1;
            xmlCloseDocument(_xmlDoc);
            unExportTree_indentXmlFile(_path,tempPath);
            remove(tempPath);
        } else {
            result = xmlDocumentToFile(_xmlDoc,tempPath) != -1;
            xmlCloseDocument(_xmlDoc);
        }
        FwException::assert(result,"Could not write xml to file: " + tempPath);
    }

    public void removeIfEmpty(const mixed nodeId){
        if ( xmlFirstChild(_xmlDoc,nodeId) < 0 ){
            xmlRemoveNode(_xmlDoc, nodeId);
            if(_currentTop == nodeId){
                this.resetTop();
            }
        }
    }

};