/**
  (c) Copyright CERN 2005. All rights not expressly granted are reserved.
  icecontrols.support@cern.ch

  SPDX-License-Identifier: LGPL-3.0-only
**/

const string TREE_WIDGET_CURRENT = "Current";
const string TREE_WIDGET_FILE = "File";
const string TREE_WIDGET_PREVIEW = "Preview";
const string TREE_WIDGET_OPERATE = "Operate";
const string TREE_WIDGET_CONFIGURE = "Configure";

const string TREE_TYPE_TREND = "Trend";
const string TREE_TYPE_WINDOW = "Window";

const int NODE_STATUS_N_A = 0;
const int NODE_STATUS_OK = 1;
const int NODE_STATUS_ERR = 2;
const int NODE_STATUS_NEW = 3;
const int NODE_STATUS_DIFF = 4;

const string NODE_PARAMNAME_PARENT = "parentname";
const string NODE_PARAMNAME_CHILDREN = "children";
const string NODE_PARAMNAME_USERDATA = "userdata";
const string NODE_PARAMNAME_DEVICEDATA = "devicedata";
const string NODE_PARAMNAME_DEVICECONFIGURATION = "deviceconfiguration";
const string NODE_PARAMNAME_NAME = "name";
const string NODE_PARAMNAME_PANELNAME = "panelname";
const string NODE_PARAMNAME_NAVIGATION = "devicenavigation";

enum UnTreeProcessorParam{
  KEEP_LOCAL_SYSTEM, // if original system name should be kept
  SYS_NAME_TO_REPLACE, // if not then by what it should be replaced
  SILENT_MODE, // if during the process user should be prompted to confirm etc
  TREE_NAME, // tree name to be put in exported file
  TREE_ROOT, // tree root node name
  SKIP_TREE_DEF, // skipping additional parameters about nodes relations
  SPARSE, // skips empty string and default values, makes a file more readable but harder to edit
  TREE_TYPE, // exported as parameter to type
  FORMAT, // format to which tree should be exported in lower case - eg. xml or json
  FORMAT_VERSION, // indicates version of serializing algorithm
  OVERWRITE_OK, // if it is ok to overwrite if file already exists
  FORCE_CONTINUE, // try to force continue import/export even if errors are encountered
  BEAUTIFY, // if output should be parsed for better readability
  DELAY, // delay in ms between processing next nodes - for debugging purposes
  NODES_STATUSES, // internal use - for listing nodes statuses
  EXPORTED_DEVICES, // internal use - for listing exported devices
  MISSING_CANDIDATES // internal use - for tracking possible missing nodes
};

const mapping TREE_PROCESSOR_PARAMS_DESCRIPTIONS = makeMapping(
  UnTreeProcessorParam::KEEP_LOCAL_SYSTEM, "If original system name should be kept",
  UnTreeProcessorParam::SYS_NAME_TO_REPLACE, "If system name should be replaced then by what",
  UnTreeProcessorParam::SILENT_MODE, "If during the process user should be prompted to confirm etc",
  UnTreeProcessorParam::TREE_NAME, "Tree name to be put in exported file",
  UnTreeProcessorParam::TREE_ROOT, "Tree root node name",
  UnTreeProcessorParam::SKIP_TREE_DEF, "Skipping additional parameters about nodes relations",
  UnTreeProcessorParam::SPARSE, "Skips empty string and default values, makes a file more readable but harder to edit",
  UnTreeProcessorParam::TREE_TYPE, "Exported as parameter to type",
  UnTreeProcessorParam::FORMAT, "Format to which tree should be exported in lower case - eg. xml or json",
  UnTreeProcessorParam::FORMAT_VERSION, "Indicates version of serializing algorithm",
  UnTreeProcessorParam::OVERWRITE_OK, "If it is ok to overwrite if file already exists",
  UnTreeProcessorParam::FORCE_CONTINUE, "Try to force continue import/export even if errors are encountered",
  UnTreeProcessorParam::BEAUTIFY, "If output should be parsed for better readability",
  UnTreeProcessorParam::DELAY, "For debuging purposes - adds delay between processing next nodes, in ms",
  UnTreeProcessorParam::NODES_STATUSES, "internal use - for listing nodes statuses",
  UnTreeProcessorParam::EXPORTED_DEVICES, "Internal use - for listing exported devices",
  UnTreeProcessorParam::MISSING_CANDIDATES, "Internal use - for tracking possible missing nodes"
);

class UnTreeProcessContext{
    #event sendLogMessage(const string msg, const string category)

    protected shared_ptr<mapping> _params;
    protected bool _hasWarnings = false;

    public UnTreeProcessContext(const mapping params = makeMapping()){
        _params = new mapping();
        this.fromMapping(params);
    }

    public mapping getDefaults(){
        return makeMapping(
            UnTreeProcessorParam::KEEP_LOCAL_SYSTEM, true,
            UnTreeProcessorParam::SYS_NAME_TO_REPLACE, getSystemName(),
            UnTreeProcessorParam::SILENT_MODE, true,
            UnTreeProcessorParam::TREE_NAME, "",
            UnTreeProcessorParam::TREE_ROOT, "",
            UnTreeProcessorParam::SKIP_TREE_DEF, false,
            UnTreeProcessorParam::SPARSE, false,
            UnTreeProcessorParam::TREE_TYPE, "Window",
            UnTreeProcessorParam::FORMAT, "xml",
            UnTreeProcessorParam::OVERWRITE_OK, true,
            UnTreeProcessorParam::FORMAT_VERSION, "2",
            UnTreeProcessorParam::OVERWRITE_OK, false,
            UnTreeProcessorParam::FORCE_CONTINUE, false,
            UnTreeProcessorParam::BEAUTIFY, true,
            UnTreeProcessorParam::DELAY, 0,
            UnTreeProcessorParam::NODES_STATUSES, makeMapping(),
            UnTreeProcessorParam::EXPORTED_DEVICES, makeDynString(),
            UnTreeProcessorParam::MISSING_CANDIDATES, makeDynString()
        );
    }

    public anytype checkValue(UnTreeProcessorParam key, anytype checkValue, anytype result){
        if(!_params.contains(key)){
            return result;
        }
        return _params[key] == checkValue ? result : _params[key];
    }

    public void fromMapping(const mapping input){
        _params.clear();
        _params = getDefaults();
        _params.merge(input);
    }

    public anytype get(UnTreeProcessorParam key){
        return _params.value(key,0);
    }

    public shared_ptr<mapping> asMap(){
        return _params;
    }

    public void set(UnTreeProcessorParam key, anytype value){
        _params[key] = value;
    }

    public void mergeParams(const mapping input){
        _params.merge(input);
    }

    public void addDevice(UnTreeProcessorParam key, string device){
        if(device == ""){
            return;
        }
        if(!_params.contains(key)){
            _params[key] = makeDynString(device);
        } else {
            _params[key].append(device);
        }
    }

    public void setStatus(const string nodeId, const anytype status){
        _params[UnTreeProcessorParam::NODES_STATUSES][nodeId] = status;
    }

    public string encodeNodeName(string nodeName){
        string result = encodeDeviceName(nodeName);
        strreplace(result,"#local:","");
        if(result.startsWith("fwTN_") || result.startsWith("unTN_")){
            result = result.right(result.length()-5);
        } 
        return result;
    }

    public string encodeDeviceName(string device){
        if(device == ""){
            return "";
        }
        const string rep = _params.value(UnTreeProcessorParam::SYS_NAME_TO_REPLACE,"");
        if(rep != "" && unExportTree_getSystemName(device) == rep){
            return "#local:" + unExportTree_removeSystemName(device);
        }
        return device;
    }

    public string decodeDeviceName(string device){
        if(device == ""){
            return "";
        }
        const string rep = _params.value(UnTreeProcessorParam::SYS_NAME_TO_REPLACE,"");
        strreplace(device,"#local:",rep);
        return device;
    }

  public void rethrowIfNeeded(){
      if(this.get(UnTreeProcessorParam::FORCE_CONTINUE)){
        sendLogMessage(FwException::last().getText(),"WARNING");
        _hasWarnings = true;
      } else {
        FwException::rethrow();
      }
  }

  public bool assertIfNeeded(const bool condition, const string msg){
    if(!condition){
      throwIfNeeded(msg);
    }
    return condition;
  }

  public void throwIfNeeded(const string msg){
      if(this.get(UnTreeProcessorParam::FORCE_CONTINUE)){
        sendLogMessage(msg,"WARNING");
        _hasWarnings = true;
      } else {
        FwException::raise(msg);
      }
  }

};