/**
  (c) Copyright CERN 2005. All rights not expressly granted are reserved.
icecontrols.support@cern.ch

  SPDX-License-Identifier: LGPL-3.0-only
**/

#uses "fwUnitTestComponentAsserts.ctl"
#uses "classes/unTree/unTreeProcessor.ctl"
#uses "classes/unTree/unTreeRepositoryImpl.ctl"

mapping processParams;
FwTreeNodePtr workingRoot = nullptr;
FwTreeNodePtr readRoot = nullptr;
const string exampleFile = getPath(DATA_REL_PATH, "test/unTree/export_simple_window.txt");
string tempFile = exampleFile;
const string testRootNodeName = "UN_TREE_UNIT_TEST_MAIN_NODE";

void test_unTree_export_setupSuite(){
	tempFile.replace("export_simple_window","export_for_unit_tests");
	// fwTreeDisplay_createTree(treeName, 2, treeName, makeDynString(), makeDynString());
	workingRoot = FwTree_Repository::getDefault().get("fwTN_"+testRootNodeName);
	if(workingRoot == nullptr){
		FwTreeNodePtr topRoot = FwTree_Repository::getDefault().get("fwTN_WindowTree");
		workingRoot = FwTreeNode::create(testRootNodeName,topRoot);
	}

	readRoot = UnTree_RepositoryReadImpl::getInstance().createRootNode("UNIT_TESTS_READ_ROOT");
	processParams = makeMapping(
		UnTreeProcessorParam::KEEP_LOCAL_SYSTEM, false,
		UnTreeProcessorParam::SKIP_TREE_DEF, false,
		UnTreeProcessorParam::FORCE_CONTINUE, true,
		UnTreeProcessorParam::TREE_TYPE, "Window",
		UnTreeProcessorParam::TREE_ROOT, "WindowTree",
		UnTreeProcessorParam::SPARSE, false,
		UnTreeProcessorParam::FORMAT, "xml",
		UnTreeProcessorParam::FORMAT_VERSION, "2",
		UnTreeProcessorParam::OVERWRITE_OK, true,
		UnTreeProcessorParam::BEAUTIFY, true,
		UnTreeProcessorParam::DELAY, 0,
		UnTreeProcessorParam::TREE_NAME, "",
		UnTreeProcessorParam::SILENT_MODE, true,
		UnTreeProcessorParam::SYS_NAME_TO_REPLACE, getSystemName()
	);
}

void test_unTree_export_teardownSuite(){
	workingRoot.removeMe(true);
}
//------------------------------------------------------------------

void logFcn(const string msg, const string tp){
	DebugTN(tp,msg);
}

void doExport(shared_ptr<UnTreeProcessor> processor, mapping params, const string filePath, FwTreeNodePtr node){
	processor.processAction("export",node,filePath,params);
}

void doReadAndImport(shared_ptr<UnTreeProcessor> processor, mapping params, const string filePath){
	processor.processAction("read",readRoot,filePath,params);
	processor.processAction("import",readRoot,"",params,workingRoot);
}

void clearWorkingTree(){
	FwTreeNodePVec kids =  workingRoot.getChildren();
	for(int i=0; i<kids.count(); ++i){
		kids.at(i).removeMe(true);
	}
	delay(0,200);
}

void test_export(){
	shared_ptr<UnTreeProcessor> processor = new UnTreeProcessor();
	classConnect("logFcn",processor,UnTreeProcessor::sendLogMessage);
	FwTreeNodePtr topRoot = FwTree_Repository::getDefault().get("fwTN_WindowTree");
	doExport(processor, processParams, tempFile, topRoot);
	string content;
	assertTrue(fileToString(tempFile, content), "reading exported file");
	DebugTN(content);
}

void test_import(){
	UnTreeProcessor::draw(workingRoot);
	shared_ptr<UnTreeProcessor> processor = new UnTreeProcessor();
	classConnect("logFcn",processor,UnTreeProcessor::sendLogMessage);
	doReadAndImport(processor, processParams, exampleFile);
	UnTreeProcessor::draw(workingRoot);
}

void test_export_clear_import(){
	clearWorkingTree();
	UnTreeProcessor::draw(workingRoot);
	shared_ptr<UnTreeProcessor> processor = new UnTreeProcessor();
	classConnect("logFcn",processor,UnTreeProcessor::sendLogMessage);
	doReadAndImport(processor, processParams, exampleFile);
	UnTreeProcessor::draw(readRoot);
	UnTreeProcessor::draw(workingRoot);
	delay(1);
	doExport(processor, processParams, tempFile, workingRoot);
	string content1, content2;
	assertTrue(fileToString(exampleFile, content1), "reading example file");
	assertTrue(fileToString(tempFile, content2), "reading exported file");
	mapping diffs = getFilesDiffs(content1,content2);
	const dyn_int lines = mappingKeys(diffs);
	for(int i=0; i<lines.count(); ++i){
		const int n = lines.at(i);
		if(diffs[n].at(0).contains(testRootNodeName) || diffs[n].at(1).contains(testRootNodeName) ){
			mappingRemove(diffs,n);
		}
	}
	if(diffs.count()>0){
		displayFilesDiffs(content1,content2,diffs);
		// assertTrue(false,"files different");
	}
}

mapping getFilesDiffs(string &content1, string &content2){
	mapping diffs;
	dyn_string f1 = strsplit(content1,"\n");
	dyn_string f2 = strsplit(content2,"\n");
	const int lim = f1.count() < f2.count() ? f1.count(0) : f2.count();
	int cnt = 0;
	for(int i=0; i<lim; ++i){
		if(f1.at(i) != f2.at(i)){
			diffs[i] = makeDynString(f1.at(i),f2.at(i));
		}
	}
	if(f1.count() != f2.count()){
		diffs[-1] = makeDynString(f1.count(),f2.count(0));
	}
	return diffs;
}

void displayFilesDiffs(string &content1, string &content2, const mapping &diffs){
	DebugN("--------------------");
	DebugN(content1);
	DebugN("--------------------");
	DebugN(content2);
	DebugN("--------------------");
	for(int i=0; i<diffs.count(); ++i){
		const int line = diffs.keyAt(i);
		if(line == -1){
			DebugN("Files lengths dont match (" + diffs[line].at(0) + " lines vs " + diffs[line].at(1) + ")");
			continue;
		}
		DebugN("diff in line: "+line);
		DebugN(diffs[line].at(0));
		DebugN(diffs[line].at(1));
	}
	if(diffs.count()>0){
		DebugN("In total " + diffs.count() + " line(s) were different.");
	}
}